// Copyright 2006 all rights reserved; see LICENSE file for BSD-style license

package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.Parser.Table.*;
import edu.berkeley.sbp.Sequence.Position;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/** a node in the GSS */
final class Node
    implements Invokable<Position, Boolean, Result>,
               IntegerMappable,
               GraphViz.ToGraphViz,
               Iterable<Result> {

    /** which GSS.Phase this Node belongs to */
    public  GSS.Phase phase() { return phase; }
    public Iterator<Result> iterator() { return results.iterator(); }
    public Parser.Table.State state() { return state; }

    public int toInt() { return idx; }

    public boolean merge(Result r) {
        if (results.contains(r)) return true;
        results.add(r);
        r.addChild(this);
        if (fromEmptyReduction) return false;
        state.invokeReductions(phase().getToken(), this, false, r);
        return false;
    }

    private boolean destroyed = false;
    public boolean check() {
        boolean dead = true;
        // - all nodes must have a parent
        // - non-doomed nodes must either:
        //      - be on the frontier or
        //      - have a non-doomed node closer to the frontier than themself

        if (phase.isFrontier()) dead = false;
        for(Result r : children)
            if (state.doomed) { if (r.usedByAnyNode()) dead = false; }
            else              { if (r.usedByNonDoomedNode()) dead = false; }
        dead |= results.size()==0;
        if (!dead || destroyed) return dead;
        destroyed = true;
        while(children.size()>0)
            for(Result r : children) {
                children.remove(r);
                r.destroy();
                break;
            }
        while(results.size()>0)
            for(Result r : results) {
                results.remove(r);
                r.removeChild(this);
                r.check();
                break;
            }
        return dead;
    }

    //////////////////////////////////////////////////////////////////////

    private static int node_idx = 0;
    private final int idx = node_idx++;

    private final GSS.Phase phase;
    private final Parser.Table.State state;
    private final boolean fromEmptyReduction;
    //private       FastSet<Result> results = new FastSet<Result>();
    private       HashSet<Result> results = new HashSet<Result>();
    private       HashSet<Result> children = new HashSet<Result>();
    public void removeChild(Result child) { children.remove(child); }
    public void removeResult(Result result) { results.remove(result); }
    public void addChild(Result child) { children.add(child); }

    public final void invoke(Position r, Boolean emptyProductions, Result only) {
        if (emptyProductions != (r.pos==0)) return;
        if (r.pos==0) finish(r, r.zero(phase().getLocation().createRegion(phase().getLocation())), phase());
        else          reduce(r, r.pos-1, phase(), only);
    }

    private void reduce(Position r, int pos, GSS.Phase target, Result only) {
        Forest[] holder = r.holder;
        Forest old = holder[pos];
        for(Result res : results)
            if (only == null || res == only) {
                Node child = res.parent();
                holder[pos] = res.getForest();
                if (pos>0)  child.reduce(r, pos-1, target, null);
                else new Reduction(child, r, r.rewrite(child.phase().getLocation().createRegion(target.getLocation())), target);
            }
        holder[pos] = old;
    }

    void finish(Position r, Forest forest, GSS.Phase target) {
        Parser.Table.State state0 = (Parser.Table.State)state.gotoSetNonTerminals.get(r.owner());
        Result result = new Result(forest, this, r);
        target.newNodeFromReduction(result, state0, r.pos<=0, r);
    }

    Node(GSS.Phase phase, Result result, State state, boolean fromEmptyReduction) {
        this.phase = phase;
        this.state = state;
        this.fromEmptyReduction = fromEmptyReduction;
        results.add(result);
        result.addChild(this);
        if (phase.hash.get(state, result.phase()) != null) throw new Error("severe problem!");
        phase.hash.put(state, result.phase(), this);

        for(Object s : state.also)
            phase.newNode(new Result(null, this, null), (State)s, fromEmptyReduction);

        state.invokeReductions(phase().token, this, true, null);
        if (!fromEmptyReduction)
            state.invokeReductions(phase().getToken(), this, false, null);
    }

    // GraphViz //////////////////////////////////////////////////////////////////////////////

    public GraphViz.Node toGraphViz(GraphViz gv) {
        if (results.size()==0) return null;
        if (gv.hasNode(this)) return gv.createNode(this);
        GraphViz.Node n = gv.createNode(this);
        n.label = ""+state.toStringx();
        n.shape = "rectangle";
        boolean hasparents = false;
        for(Result r : results) n.edge(r, "");
        n.color = state.doomed ? "red" : "green";
        ((GraphViz.Group)phase().toGraphViz(gv)).add(n);
        return n;
    }
    public boolean isTransparent() { return false; }
    public boolean isHidden() { return false; }

}
